/*
  Copyright (c) 2008, Arturas Moskvinas. All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License. You should have
  received a copy of the GPL license along with this program; if you
  did not, you can find it at http://www.gnu.org/
*/

#include "httpd.h"
#include "http_log.h"
#include "http_request.h"
#include "ap_mpm.h"
#include "scoreboard.h"

#include "unixd.h"
#ifndef _BSD_SOURCE 
#define _BSD_SOURCE
#endif
/* ^^^glibc compatability */
#include <stdlib.h>
#include <math.h>
module AP_MODULE_DECLARE_DATA overload_module;

static int server_limit=0;
static int thread_limit=0;

/*configuration*/
typedef struct overload_cfg {
    apr_int64_t oneminloadavg;
    apr_int64_t fiveminloadavg;
    apr_int64_t fifteenminloadavg;
    apr_int64_t enabled;
    apr_int64_t log_requests;
    apr_int64_t kill_requests;
    apr_int64_t lower_precedence;
    apr_int64_t allow_503_page;
} overload_cfg;
/*load avg, just in case if some OS specific will be needed*/
typedef struct loadavg {
        apr_off_t one_minute;
        apr_off_t five_minute;
        apr_off_t fifteen_minute;
} loadavg;


static void* overload_dconf_create(apr_pool_t *p, char * dir)
{
    overload_cfg *cfg;
    cfg = (overload_cfg *) apr_palloc(p, sizeof(overload_cfg));
    cfg->oneminloadavg = -1; /*defaut values (ignore)*/
    cfg->fiveminloadavg = -1;
    cfg->fifteenminloadavg = -1;
    cfg->enabled = -1; /*not enabled by default*/
    cfg->log_requests = -1; /*not enabled by default*/
    cfg->kill_requests = -1;
    cfg->lower_precedence = -1;
    cfg->allow_503_page = -1;
    ap_mpm_query(AP_MPMQ_HARD_LIMIT_THREADS, &thread_limit);
    ap_mpm_query(AP_MPMQ_HARD_LIMIT_DAEMONS, &server_limit);
    return (void *)cfg;
}

static void *overload_dconf_merge(apr_pool_t *p, void *conf1, void *conf2)
{
    /*values are inherited from parent if it is not set in current context*/
    overload_cfg *cfg;
    cfg = (overload_cfg *) apr_palloc(p, sizeof(overload_cfg));
    overload_cfg *over_conf1 = (overload_cfg *) conf1;
    overload_cfg *over_conf2 = (overload_cfg *) conf2; 
    cfg->oneminloadavg=over_conf2->oneminloadavg!=-1 ? over_conf2->oneminloadavg : over_conf1->oneminloadavg;
    cfg->fiveminloadavg=over_conf2->fiveminloadavg!=-1 ? over_conf2->fiveminloadavg : over_conf1->fiveminloadavg;
    cfg->fifteenminloadavg=over_conf2->fifteenminloadavg!=-1 ? over_conf2->fifteenminloadavg : over_conf1->fifteenminloadavg;
    cfg->enabled=over_conf2->enabled!=-1 ? over_conf2->enabled : over_conf1->enabled;
    cfg->log_requests=over_conf2->log_requests!=-1 ? over_conf2->log_requests : over_conf1->log_requests;
    cfg->kill_requests=over_conf2->kill_requests!=-1 ? over_conf2->kill_requests : over_conf1->kill_requests;
    cfg->lower_precedence=over_conf2->lower_precedence!=-1 ? over_conf2->lower_precedence : over_conf1->lower_precedence;
    cfg->allow_503_page=over_conf2->allow_503_page!=-1 ? over_conf2->allow_503_page : over_conf1->allow_503_page;
    return (void *)cfg;
}

static const char* set1min(cmd_parms *cmd, void *mconfig, const char *arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->oneminloadavg=apr_atoi64(arg1);
    return NULL;
}

static const char* set5min(cmd_parms *cmd, void *mconfig, const char *arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->fiveminloadavg=apr_atoi64(arg1);
    return NULL;
}

static const char* set15min(cmd_parms *cmd, void *mconfig, const char *arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->fifteenminloadavg=apr_atoi64(arg1);
    return NULL;
}


static const char* setenableflag(cmd_parms *cmd, void *mconfig, const int arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->enabled=arg1;
    return NULL;
}

static const char* setlogflag(cmd_parms *cmd, void *mconfig, const int arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->log_requests=arg1;
    return NULL;
}

static const char* setprecedenceflag(cmd_parms *cmd, void *mconfig, const int arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->lower_precedence=arg1;
    return NULL;
}

static const char* setkillflag(cmd_parms *cmd, void *mconfig, const int arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->kill_requests=arg1;
    return NULL;
}

static const char* set503flag(cmd_parms *cmd, void *mconfig, const int arg1)
{
    overload_cfg *cfg = mconfig;
    cfg->allow_503_page=arg1;
    return NULL;
}

/*Log currently running process */
static void log_requests(request_rec *r)
{
    worker_score *w_record;
    int i, j;
    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server, "mod_overload: Currently running requests:");
    for (i=0; i < server_limit; i++) //daemons
    {
        for (j=0; j < thread_limit; j++)//threads
        {
    	    w_record = ap_get_scoreboard_worker(i,j);
	    if ((w_record->access_count==0) && ((w_record->status==SERVER_READY) || (w_record->status==SERVER_DEAD))) /*not running workers*/
	    {
	        continue;
	    }
	    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server, "mod_overload[%li]: Hostname:%s, uri: %s, client: %s", r->connection->id, w_record->vhost, w_record->request, w_record->client);
	}
    }
}
/*Kill all currently running processes, including the one which started killing process, not elegant solution*/
static void kill_requests(request_rec *r)
{/*TODO: add code to kill all request except the one which started kill process*/
    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server, "mod_overload: Killing running requests:");
    ap_os_killpg(0, 9);
}



static int mod_overload_method_handler (request_rec *r)
{
    double ldavg[3];
    server_rec *s=r->server;
    loadavg avg;
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_overload: %s", "Handling request");
    overload_cfg *cfg=(overload_cfg*) ap_get_module_config(r->per_dir_config, &overload_module);

    /*no config or not enabled*/
    if ((cfg==NULL) || (cfg->enabled==0))
    {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_overload: %s", "Doing nothing, Overload not enabled");
	return DECLINED;
    }
    
    /*check if it is internal process which is looking for 503 error page, if it is, let it through, or else we won't have custom 504 pages*/
    if ((cfg->allow_503_page==1) && (r->status==HTTP_SERVICE_UNAVAILABLE))
    {
	ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_overload: %s", "Request status is 503, then proceed (to display ErrorDocument)");
	return DECLINED;
    }

    avg.one_minute=-1;
    avg.five_minute=-1;
    avg.fifteen_minute=-1;
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, "mod_overload: %s", "Getting load averages (OS specific functions)");
    if (getloadavg(ldavg, 3)==-1)
    {
	ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,"mod_overload: %s", "Unable to get load average");
	return DECLINED;
    }
    avg.one_minute=floor(ldavg[0]);
    avg.five_minute=floor(ldavg[1]);
    avg.fifteen_minute=floor(ldavg[2]);
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_overload: Average load: 1min %i, 5min %i, 15min %i", avg.one_minute, avg.five_minute, avg.fifteen_minute);
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_overload: Configuration load: 1min %i, 5min %i, 15min %i", cfg->oneminloadavg, cfg->fiveminloadavg, cfg->fifteenminloadavg);

    if ((cfg->oneminloadavg!=-1) && (cfg->oneminloadavg<avg.one_minute))
    {
        ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,"mod_overload: Overloaded, 1 minute load average, current load: %i, max load allowed: %i", avg.one_minute, cfg->oneminloadavg);
	if (cfg->log_requests>0)
	{
	    log_requests(r);
	}
	if (cfg->kill_requests>0)
	{
	    kill_requests(r);
	}
	return HTTP_SERVICE_UNAVAILABLE;
    }

    if ((cfg->fiveminloadavg!=-1) && (cfg->fiveminloadavg<avg.five_minute))
    {	
	/* if 1min load average less then 5min, then skip stopping, overload is over, else proceed */
	if ((cfg->lower_precedence<1) || ((cfg->lower_precedence>0) && (avg.one_minute>=avg.five_minute))) 
        {
	    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,"mod_overload: Overloaded, 5 minute load average, current load: %i, max load allowed: %i", avg.five_minute, cfg->fiveminloadavg);
	    if (cfg->log_requests>0)
	    {
		log_requests(r);
	    }
	    if (cfg->kill_requests>0)
	    {
		kill_requests(r);
	    }
	    return HTTP_SERVICE_UNAVAILABLE;
	}
	else
	{
	    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,"mod_overload: Overloaded, 1 minute load is lower than 5minute, skipping, 1min load %i, 5min load %i, configured 5min load %i", avg.one_minute, avg.five_minute, cfg->fiveminloadavg);
	}
    }
    if ((cfg->fifteenminloadavg!=-1) && (cfg->fifteenminloadavg<avg.fifteen_minute))
    {
	/* if 5min load average less then 15min, then skip stopping, overload is over, else proceed */
        if ((cfg->lower_precedence<1) || ((cfg->lower_precedence>0) && (avg.five_minute>=avg.fifteen_minute)))
        {
	    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,"mod_overload: Overloaded, 15 minute load average, current load: %i, max load allowed: %i", avg.fifteen_minute, cfg->fifteenminloadavg);
	    if (cfg->log_requests>0)
	    {
		log_requests(r);
	    }
	    if (cfg->kill_requests>0)
	    {
		kill_requests(r);
	    }
	    return HTTP_SERVICE_UNAVAILABLE;
	}
	else
	{
	    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,"mod_overload: Overloaded, 5 minute load is lower than 15 minute, skipping, 5min load %i, 15min load %i, configured 15 min load %i", avg.five_minute, avg.fifteen_minute, cfg->fifteenminloadavg);
	}
    }/*we checked load avg, so it is OK*/
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"mod_overload: %s", "Not overloaded, returning OK");
    return DECLINED;
}

static void mod_overload_register_hooks (apr_pool_t *p)
{
  ap_hook_handler(mod_overload_method_handler, NULL, NULL, APR_HOOK_FIRST);
}

static const command_rec overload_cmds[] = {
AP_INIT_TAKE1(
"OverloadLoadAvg1min",
set1min,
NULL,
RSRC_CONF|ACCESS_CONF,
"1 min load average, when request is stopped"),

AP_INIT_TAKE1(
"OverloadLoadAvg5min",
set5min,
NULL,
RSRC_CONF|ACCESS_CONF,
"5 min load average, when request is stopped"),

AP_INIT_TAKE1(
"OverloadLoadAvg15min",
set15min,
NULL,
RSRC_CONF|ACCESS_CONF,
"15 min load average, when request is stopped"),

AP_INIT_FLAG(
"OverloadEnable",
setenableflag,
NULL,
RSRC_CONF|ACCESS_CONF,
"Enable OverloadModule"),

AP_INIT_FLAG(
"OverloadLogRequests",
setlogflag,
NULL,
RSRC_CONF|ACCESS_CONF,
"Log Request which where running after Overload detected load avg too high"),

AP_INIT_FLAG(
"OverloadKillAllRequests",
setkillflag,
NULL,
RSRC_CONF|ACCESS_CONF,
"Kill all currently running requests (not recommended)"),

AP_INIT_FLAG(
"OverloadAllow503Request",
set503flag,
NULL,
RSRC_CONF|ACCESS_CONF,
"During overload allow any pages which have 503 status code to proceed (allow custom ErrorPage)"),

AP_INIT_FLAG(
"OverloadLowerHasPrecedence",
setprecedenceflag,
NULL,
RSRC_CONF|ACCESS_CONF,
"Do not stop request if 1min load is lower than 5 min load (effect after overloading, server sits idling for too long)"),
{NULL}
};


/*--------------------------------------------------------------------------*/
/*                                                                          */
/* Data declarations.                                                       */
/*                                                                          */
/* Here are the static cells and structure declarations private to our      */
/* module.                                                                  */
/*                                                                          */
/*--------------------------------------------------------------------------*/
module AP_MODULE_DECLARE_DATA overload_module =
{
    STANDARD20_MODULE_STUFF,
    overload_dconf_create,    /* per-directory config creator */
    overload_dconf_merge,     /* rules are the same as per dir... */
    NULL,                     /* server config creator */
    NULL,                     /* server config merger */
    overload_cmds,              	/* command table */
    mod_overload_register_hooks,       /* callback for registering hooks */
};
